import os
import shutil
import sys
import unittest

class TestProblem(unittest.TestCase):

    def setUp(self) -> None:
        self.cwd = os.getcwd() 
        self.folder = "/tmp/IOHexperimenter"
        return super().setUp()

    def tearDown(self) -> None:
        shutil.rmtree(self.folder, True)
        os.chdir(self.cwd)
        return super().tearDown()

    @unittest.skipUnless(sys.platform != "win32", "windows tmp dir")
    def test_downloader(self):
        self.assertFalse(os.path.isdir(self.folder))
        os.chdir("/tmp")
        import ioh
        self.assertTrue(any(ioh.ProblemClass.GRAPH.problems))


if __name__ == "__main__":
    unittest.main()


from typing import Any, ClassVar, Enum

MAX: OptimizationType
MIN: OptimizationType

class AbstractIntegerConstraint:
    enforced: Any
    weight: float

    def __call__(self, *args, **kwargs) -> Any: ...
    def is_feasible(self, *args, **kwargs) -> Any: ...
    def compute_violation(self, *args, **kwargs) -> Any: ...
    def penalize(self, *args, **kwargs) -> Any: ...
    def violation(self, *args, **kwargs) -> Any: ...
    def penalty(self, *args, **kwargs) -> Any: ...

class AbstractRealConstraint:
    enforced: Any
    weight: float

    def __call__(self, *args, **kwargs) -> Any: ...
    def is_feasible(self, *args, **kwargs) -> Any: ...
    def compute_violation(self, *args, **kwargs) -> Any: ...
    def penalize(self, *args, **kwargs) -> Any: ...
    def violation(self, *args, **kwargs) -> Any: ...
    def penalty(self, *args, **kwargs) -> Any: ...

class IntegerConstraint(AbstractIntegerConstraint):
    def __init__(self, *args, **kwargs) -> None: ...

class RealConstraint(AbstractRealConstraint):
    def __init__(self, *args, **kwargs) -> None: ...

class IntegerBounds:
    def __init__(self, size: int, lb: int, ub: int) -> None: ...
    def compute_violation(self, *args, **kwargs) -> Any: ...
    @property
    def lb(self) -> Any: ...
    @property
    def ub(self) -> Any: ...

class IntegerSolution:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def x(self) -> Any: ...
    @property
    def y(self) -> Any: ...

class IntegerState:
    def __init__(self, initial: IntegerSolution) -> None: ...
    @property
    def current(self) -> IntegerSolution: ...
    @property
    def current_best(self) -> IntegerSolution: ...
    @property
    def current_best_internal(self) -> IntegerSolution: ...
    @property
    def current_internal(self) -> IntegerSolution: ...
    @property
    def evaluations(self) -> int: ...
    @property
    def optimum_found(self) -> bool: ...

class LogInfo:
    def __init__(
        self,
        evaluations: int,
        raw_y_best: float,
        transformed_y: float,
        transformed_y_best: float,
        current: RealSolution,
        optimum: RealSolution,
    ) -> None: ...
    @property
    def current(self) -> RealSolution: ...
    @property
    def evaluations(self) -> int: ...
    @property
    def objective(self) -> RealSolution: ...
    @property
    def transformed_y(self) -> float: ...
    @property
    def transformed_y_best(self) -> float: ...
    @property
    def y_best(self) -> float: ...

class MetaData:
    def __init__(
        self,
        problem_id: int,
        instance: int,
        name: str,
        n_variables: int,
        optimization_type: OptimizationType,
    ) -> None: ...
    @property
    def instance(self) -> int: ...
    @property
    def n_variables(self) -> int: ...
    @property
    def name(self) -> str: ...
    @property
    def optimization_type(self) -> Any: ...
    @property
    def problem_id(self) -> int: ...

class OptimizationType:
    __doc__: ClassVar[str] = ...  # read-only
    __members__: ClassVar[dict] = ...  # read-only
    MAX: ClassVar[OptimizationType] = ...
    MIN: ClassVar[OptimizationType] = ...
    __entries: ClassVar[dict] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class RealBounds:
    def __init__(self, size: int, lb: float, ub: float) -> None: ...
    def compute_violation(self, *args, **kwargs) -> Any: ...
    @property
    def lb(self) -> Any: ...
    @property
    def ub(self) -> Any: ...

class RealSolution:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def x(self) -> Any: ...
    @property
    def y(self) -> Any: ...

class RealState:
    def __init__(self, initial: RealSolution) -> None: ...
    @property
    def current(self) -> RealSolution: ...
    @property
    def current_best(self) -> RealSolution: ...
    @property
    def current_best_internal(self) -> RealSolution: ...
    @property
    def current_internal(self) -> RealSolution: ...
    @property
    def evaluations(self) -> int: ...
    @property
    def optimum_found(self) -> bool: ...



import os
import traceback
import json
import sys
import shutil
import unittest
import io
from contextlib import redirect_stdout
from functools import partial

BASE_DIR = os.path.realpath(os.path.join(os.path.dirname(__file__), "../../"))
GB = globals()
LC = locals()


def iter_notebook(filename):
    with open(filename) as fp:
        nb = json.load(fp)

    for i, cell in enumerate(nb["cells"], 1):
        if cell["cell_type"] == "code":
            source = "".join(
                line for line in cell["source"] if not line.startswith("%")
            )
            yield i, source


def test_notebook_runner(self, notebook):
    assert os.path.isfile(notebook)
    for i, block in iter_notebook(notebook):
        with io.StringIO() as buf, redirect_stdout(buf):
            try:
                exec(block, GB, LC)
            except Exception as e:
                assert (
                    False
                ), f"failed in {notebook} cell {i}.\n\nReason:\n{e}.\n\nBlock:\n{block}"


class MetaTest(type):
    def __new__(cls, name, bases, dct):
        instance = super().__new__(cls, name, bases, dct)
        dirname = os.path.normpath(os.path.join(BASE_DIR, "example"))
        for f in filter(lambda x: x.endswith("ipynb"), os.listdir(dirname)):
            fname, *_ = os.path.basename(f).split(".")
            notebook = os.path.join(dirname, f)
            setattr(
                instance,
                f"test_notebook_{fname}",
                partial(test_notebook_runner, instance, notebook),
            )
        return instance


class TestExamples(unittest.TestCase, metaclass=MetaTest):

    """Examples test"""

    @unittest.skipUnless(sys.version_info.minor >= 7, "python version > 3.7")
    def test_python_readme(self):
        try:
            fname = os.path.join(BASE_DIR, "ioh", "README.md")
            self.assertTrue(os.path.isfile(fname))
            with open(fname) as f:
                data = f.read().split("```")
                with io.StringIO() as buf, redirect_stdout(buf):
                    for i, x in enumerate(data):
                        if x.startswith("python"):
                            block = x[6:].strip()
                            if not "help" in block:
                                try:
                                    exec(block, GB, LC)
                                except Exception as e:
                                    raise Exception(
                                        f"failed in cell {i}. Reasion {e}.\n{traceback.format_exc()}"
                                    )
        except:
            raise
        finally:
            shutil.rmtree("temp", ignore_errors=True)
            shutil.rmtree("temp2", ignore_errors=True)
            shutil.rmtree("temp3", ignore_errors=True)
            shutil.rmtree("ioh_data", ignore_errors=True)
            if os.path.exists("ioh_data.zip"):
                os.remove("ioh_data.zip")


if __name__ == "__main__":
    unittest.main()



import unittest

import ioh


def problem(x):
    return sum(x)

def zero(x):
    return 0.


def tx(x, iid):
    x[1] = iid
    return x


def ty(y, iid):
    return y * iid


def co(iid, dim):
    return [iid] * dim, iid


class TestWrappedProblem(unittest.TestCase):
    def test_wrap_problem(self):
        p = ioh.wrap_problem(
            problem,
            "problem",
            ioh.ProblemClass.REAL,
            dimension=2,
            transform_variables=tx,
            transform_objectives=ty,
            calculate_objective=co,
        )
        self.assertEqual(p([1, 2]), 2.0)
        self.assertEqual(p.optimum.x[0], 1)
        self.assertEqual(p.optimum.x[0], p.optimum.y)
    
    # @unittest.skip("can cause a segfault (cpp exit handler)")
    def test_wrap_problem_scoped(self):
        def w():
            ioh.problem.wrap_real_problem(lambda _: 0.0, "l")
            return ioh.get_problem("l")
        p = w()
        y = p([0]*5)
        self.assertEqual(y, 0.0)

    def test_wrap_problem_lambda(self):
        l = lambda _: 0.0
        ioh.problem.wrap_real_problem(l, "f")
        p = ioh.get_problem("f")
        p([0]*5)

        ioh.problem.wrap_real_problem(l, "f")
        p2 = ioh.get_problem("f")
        self.assertEqual(p.meta_data.problem_id, p2.meta_data.problem_id)
        self.assertEqual(p.meta_data.name, p2.meta_data.name)

    def test_wrap_problem_builtins(self):
        for f in (sum, min, max):
            ioh.problem.wrap_real_problem(f, f.__name__)
            p = ioh.get_problem(f.__name__)
            y = p([0]*5)
            self.assertEqual(y, 0.0)

    def test_wrap_problem_constrains(self):
        c = ioh.RealConstraint(lambda x: float(x[0] > 1), 10.0)
        p = ioh.wrap_problem(zero, "test", ioh.ProblemClass.REAL, dimension = 2, constraints=[c])
        self.assertEqual(p([10, 0]), 10.0)
        p.remove_constraint(c)
        self.assertEqual(p([10, 0]), 0)


if __name__ == "__main__":
    unittest.main()



from typing import Iterator, List

import ioh.iohcpp.logger

class IntegerBase:
    def __init__(self, ids: List[int], iids: List[int], dims: List[int]) -> None: ...
    def attach_logger(self, logger: ioh.iohcpp.logger.Logger) -> None: ...
    def detach_logger(self) -> None: ...
    def reset(self) -> None: ...
    def __iter__(self) -> Iterator: ...
    def __len__(self) -> int: ...
    @property
    def dimensions(self) -> List[int]: ...
    @property
    def instances(self) -> List[int]: ...
    @property
    def name(self) -> str: ...
    @property
    def problem_ids(self) -> List[int]: ...

class RealBase:
    def __init__(self, ids: List[int], iids: List[int], dims: List[int]) -> None: ...
    def attach_logger(self, arg0: ioh.iohcpp.logger.Logger) -> None: ...
    def detach_logger(self) -> None: ...
    def reset(self) -> None: ...
    def __iter__(self) -> Iterator: ...
    def __len__(self) -> int: ...
    @property
    def dimensions(self) -> List[int]: ...
    @property
    def instances(self) -> List[int]: ...
    @property
    def name(self) -> str: ...
    @property
    def problem_ids(self) -> List[int]: ...


class Real(RealBase): ...
class BBOB(RealBase): ...
class SBOX(RealBase): ...
class RealStarDiscrepancy(RealBase): ...
class Integer(IntegerBase): ...
class PBO(IntegerBase): ...
class Submodular(IntegerBase): ...
class IntegerStarDiscrepancy(RealBase): ...


import os
import collections
import subprocess
import shutil
import warnings
import xmltodict

BASE_DIR = os.path.realpath(os.path.dirname(os.path.dirname(__file__)))

unit_template = '''{name}
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
.. doxygen{dtype}:: {ns}::{name}
'''

namespace_template = '''
Namespaces
------------------------------------------------------------------------------------------------
.. toctree::
    :maxdepth: 1

    {subns}
'''

template = '''
{name} 
================================================================================================
{namespaces}
{types}{classes}{structs}{enums}{functions}{variables}
'''

def make_template(nsdata):
    ns = nsdata['name']
    classes = nsdata.get('classes')
    structs = nsdata.get('structs')
    functions = nsdata.get('func')
    enums = nsdata.get('enum')
    types = nsdata.get('typedef')
    variables = nsdata.get('var')
    
    if ns == 'ioh':
        nsdata['namespaces'] = None

    def formatter(dtype, title, objs):
        return '' if objs is None else f'\n{title}\n--------------------------------\n' + '\n'.join(
            unit_template.format(
                ns=ns, 
                dtype=dtype, 
                name=x.split(f"{ns}::")[-1]) for x in objs
        )    

    return template.format(
        name=ns.split("::")[-1], 
        ns=ns,
        namespaces='' if nsdata['namespaces'] is None else namespace_template.format(
            subns="\n    ".join([f'{x.replace("::", "_")}.rst' for x in nsdata['namespaces']])
        ),
        types=formatter('typedef', 'Types', types),
        classes=formatter('class', 'Classes', classes),
        structs=formatter('struct', 'Structs', structs),
        enums=formatter('enum', 'Enumerations', enums),
        functions=formatter('function', 'Functions', functions),
        variables=formatter('variable', 'Variables', variables),
    )

def get_namespace_files(xmldir):
    for f in os.listdir(xmldir):
        if not f.endswith(".xml"): continue
        if not f.startswith("namespace"): continue
        yield os.path.join(xmldir, f)

def get_values(list_of_d, attr ='#text'):
    if list_of_d is None:
        return 

    if isinstance(list_of_d, (collections.OrderedDict, dict)):
        list_of_d = [list_of_d]

    return list(sorted(set([x[attr] for x in list_of_d])))  
    
        
def generate_sphinx_templates_from_xml(xmldir = "doc/build/xml", outdir = "doc/python/source/cpp"):
    assert os.path.isdir(xmldir), xmldir
    assert os.path.isdir(outdir), outdir

    ns = {}
    for f in get_namespace_files(xmldir):
        with open(f) as h:
            data = xmltodict.parse(h.read())
        name = data['doxygen']['compounddef']['compoundname']
        namespaces = get_values(data['doxygen']['compounddef'].get('innernamespace'))
        classes = get_values(data['doxygen']['compounddef'].get('innerclass'))
        sections = data['doxygen']['compounddef'].get('sectiondef')
        nsdata = dict(name=name, namespaces=namespaces)

        
    
        if classes is not None:
            classes = []
            structs = []
            cdata = data['doxygen']['compounddef'].get('innerclass')
            if isinstance(cdata, (dict, collections.OrderedDict)):
                cdata = [cdata]
            for e in cdata:
                if e['@refid'].startswith('class'):
                    classes.append(e['#text'])
                else:
                    structs.append(e['#text'])

            nsdata['classes'] = classes if any(classes) else None
            nsdata['structs'] = structs if any(structs) else None
        
        if sections is not None:
            if isinstance(sections, (dict, collections.OrderedDict)):
                sections = [sections]
            for sec in sections:
                nsdata[sec['@kind']] = [f"{name}::{x}" for x in get_values(sec['memberdef'], "name")]
                if name == "ioh::logger" and sec['@kind'] == 'var':
                    nsdata[sec['@kind']] = [x for x in nsdata[sec['@kind']] if not "::_" in x]

        ns[name] = nsdata
        
        if name == "ioh::problem":
            nsdata['func'] = [x for x in nsdata['func'] if not "InstanceBasedProblem" in x]
        
        filename = f'{name.replace("::", "_")}.rst'
        if filename in ("fmt.rst"):
            continue
        with open(os.path.join(outdir, filename), "w+") as f:
            f.write(make_template(nsdata))


def generate_stubs(sourcedir):
    # remove any existing stubs
    for stubfile in os.listdir(os.path.join(sourcedir, "ioh")):
        stubfile = os.path.join(sourcedir, "ioh", stubfile)
        if stubfile.endswith(".pyi"):
            os.remove(stubfile)

    for subdir in ("iohcpp", "logger"):
        if os.path.isdir(os.path.join(sourcedir, "ioh", subdir)):
            shutil.rmtree(os.path.join(sourcedir, "ioh", subdir))

    # generate stub files
    command = """stubgen -m ioh -p ioh.iohcpp -o ./"""
    subprocess.check_call(command, cwd=sourcedir, shell=True)


def run_doxygen():
    out = subprocess.run(
        "cmake -DBUILD_DOCS=ON --build /home/jacob/code/IOHexperimenter/build --config Release --target doc --",
        shell=True, cwd=BASE_DIR, capture_output=True
    )
    for line in out.stdout.decode("utf-8").splitlines():
        if "warning" in line and not "clutchlog" in line:
            warnings.warn(line.replace("/home", " /home"))


def main():
    try:
        run_doxygen()
        generate_sphinx_templates_from_xml()
        directory = os.path.join(BASE_DIR, "doc", "python")
        try:
            shutil.rmtree(os.path.join(directory, "source", "api"))
        except FileNotFoundError:
            pass
        subprocess.check_call(f"make html", shell=True, cwd=directory)

        shutil.copytree(
            os.path.join(BASE_DIR, "doc", "build", "html"),
            os.path.join(BASE_DIR, "docs"),
            dirs_exist_ok=True
        )
    except Exception as reason:
        warnings.warn(f"Cannot compile docs reason: {reason}", RuntimeWarning)

if __name__ == '__main__':
    main()
        



'''Python interface of IOH package. Includes several ease-of-use routines not available in C++.'''

import os
import enum
import math
import warnings
import itertools
import multiprocessing
import typing
import shutil
import copy
import json

import urllib.request
import tarfile

 
from .iohcpp import (
    problem,
    suite, 
    logger,
    OptimizationType,
    RealSolution,
    IntegerSolution,
    IntegerBounds,
    RealBounds,
    IntegerConstraint,
    RealConstraint,
    ConstraintEnforcement,
    RealState, 
    IntegerState,
    MetaData,
    LogInfo,
)



ProblemType = typing.Union[problem.RealSingleObjective, problem.IntegerSingleObjective]
VariableType = typing.Union[int, float]
ObjectiveType = typing.List[VariableType]


def download_static_folder(warn = True):
    '''Download static files from github. 

    Required for Graph problems to be instantiated properly
    '''

    target = os.path.realpath("IOHexperimenter")
    branch = "master"

    if os.path.isdir(target) and warn:
        warnings.warn(f"Attempting to download static folder but path {target} already exists. Skipping...")
        return 
    
    os.makedirs(target)
    github_static_folder = f"https://github.com/IOHprofiler/IOHexperimenter/blob/{branch}/static.tar.gz?raw=true" 

    with urllib.request.urlopen(github_static_folder) as f:
        thetarfile = tarfile.open(fileobj=f, mode="r|gz")
        thetarfile.extractall(target)


def load_graph_problems():
    '''Helper to load Graph problems. '''

    download_static_folder(False)

    for problem_class in (
        problem.MaxCut,
        problem.MaxCoverage,
        problem.MaxInfluence,
        problem.PackWhileTravel
    ):
        problem_class.load_instances()

    assert any(problem.GraphProblem.problems), "loading the graph files failed"


class ProblemClass(enum.Enum):
    '''Enum for different problem types, values are class names.
    '''

    REAL = "RealSingleObjective"
    INTEGER = "IntegerSingleObjective"

    BBOB = "BBOB"
    STAR_REAL = "RealStarDiscrepancy"
    SBOX = "SBOX"

    PBO = "PBO"
    GRAPH = "GraphProblem"
    STAR_INTEGER = "IntegerStarDiscrepancy"

    def is_real(self):
        return self in (
                ProblemClass.REAL, 
                ProblemClass.BBOB,
                ProblemClass.SBOX,
                ProblemClass.STAR_REAL,
                ) 
    
    def is_single_objective(self):
        return True

    @property
    def problems(self):
        base_problem = getattr(problem, self.value)
        if base_problem:
            if self is ProblemClass.GRAPH and not any(base_problem.problems):
                load_graph_problems()
            return base_problem.problems


def get_problem(
    fid: typing.Union[int, str],
    instance: int = 1,
    dimension: int = 5,
    problem_class: ProblemClass = ProblemClass.REAL,
) -> ProblemType:
    '''Instantiate a problem based on its function ID, dimension, instance and suite

    Parameters
    ----------
    fid: int or str
        The function ID of the problem in the suite, or the name of the function as string
    instance: int = 1
        The instance ID of the problem
    dimension: int = 5
        The dimension (number of variables) of the problem
    problem_class: ProblemClass = ProblemClass.REAL
        The type of the problem.

    '''
    if not isinstance(problem_class, ProblemClass):
        raise AttributeError(f"problem_class should be of type {ProblemClass}")

    if ( 
        not problem_class.is_real()
        and fid in [21, 23, "IsingTriangular", "NQueens"]
    ):
        if not math.sqrt(dimension).is_integer():
            raise ValueError(
                "For this function, the dimension needs to be a perfect square!"
            )
    if (
        problem_class.is_real()
        and fid in range(1, 25)
    ):
        if not dimension >= 2:
            raise ValueError("For BBOB functions the minimal dimension is 2")

    base_problem = getattr(problem, problem_class.value)
    if base_problem:
        if fid not in (base_problem.problems.values() | base_problem.problems.keys()):
            raise ValueError(
                f"{fid} is not registered for problem type: {problem_class}"
            )
    
        if problem_class is ProblemClass.GRAPH and not any(base_problem.problems):
            load_graph_problems()
        return base_problem.create(fid, instance, dimension)

    raise ValueError(f"Problem type {problem_class} is not supported")


def wrap_problem(
    function: typing.Callable[[ObjectiveType], float],
    name: str = None,
    problem_class: ProblemClass = ProblemClass.REAL,
    dimension: int = 5,
    instance: int = 1,
    optimization_type: OptimizationType = OptimizationType.MIN,
    lb: VariableType = 0,
    ub: VariableType = 1,
    transform_variables: typing.Callable[[ObjectiveType, int], ObjectiveType] = None,
    transform_objectives: typing.Callable[[float, int], float] = None,
    calculate_objective: typing.Callable[
        [int, int], typing.Union[IntegerSolution, RealSolution]
    ] = None,
    constraints: typing.List[typing.Union[IntegerConstraint, RealConstraint]] = None

) -> ProblemType:
    '''Function to wrap a callable as an ioh function

    Parameters
    ----------
    function: fn(x: list) -> float
        The callable to wrap
    name: str = None
        The name of the function. This can be used to create new instances of this function.
        Note, when not using unique names, this will override the previously wrapped functions.
    problem_class: ProblemClass = ProblemClass.REAL
        The type of the problem.
    dimension: int = 5
        The dimension (number of variables) of the problem
    instance: int = 1
        The instance ID of the problem
    optimization_type: OptimizationType = OptimizationType.MIN
        The type of optimization to do, maximization or minimization
    lb: [int, float] = 0
        The lower bound of the constraint, should be the same type as problem_class. Defaults to 0
    ub: [int, float] = 1
        The upper bound of the constraint, should be the same type as problem_class. Defaults to 1
    transform_variables: fn(x: list) -> list = None
        A function to transform the elements of x, prior to calling the function.
    transform_objectives: fn(y: float) -> float = None
        A function to tranform the float value of y, after callling the function.
    calculate_objective: fn(instance, dimension) -> Solution | (x, y) = None
        A function to calculate the global optimum of the function. This function gets a dimension and instance id,
        and should return either a Solution objective(IntegerSolution or RealSolution) or a tuple giving the
        x and y values for the global optimum. Where x is the search space representation and y the target value.
    constraints: list[IntegerConstraint | RealConstraint] = None
        The constraints applied to the problem
    '''
    if not isinstance(problem_class, ProblemClass):
        raise AttributeError(f"problem_class should be of type {ProblemClass}")

    if name is None:
        name = function.__name__
    
    if not problem_class.is_single_objective():
        raise ValueError(
            f"Problem type {problem_class} is not supported."
        )

    wrapper = problem.wrap_real_problem if problem_class.is_real() \
        else problem.wrap_integer_problem 

    wrapper(
        function,
        name,
        optimization_type,
        lb,
        ub,
        transform_variables,
        transform_objectives,
        calculate_objective,
        [] if constraints is None else constraints
    )
    return get_problem(name, instance, dimension, problem_class)


def get_problem_id(problem_name: str, problem_class: ProblemClass) -> int: 
    '''Get the problem id corresponding to a problem name of a given type
    
    Parameters
    ----------
    problem_name: str
        The name of the problem
    problem_class: ProblemClass
        The type of the problem
    
    Returns
    -------
    int: the id of the problem
    '''    
    if not isinstance(problem_class, ProblemClass):
        raise AttributeError(f"problem_class should be of type {ProblemClass}")

    return {v: k for k, v in getattr(problem, problem_class.value).problems.items()}.get(
        problem_name
    )


class Experiment:
    '''Class to help easily setup benchmarking experiments. '''

    def __init__(
        self,
        algorithm: typing.Any,
        fids: typing.List[int],
        iids: typing.List[int],
        dims: typing.List[int],
        reps: int = 1,
        problem_class: ProblemClass = ProblemClass.REAL,
        njobs: int = 1,
        logged: bool = True,
        logger_triggers: typing.List[logger.trigger.Trigger] = [
            logger.trigger.ON_IMPROVEMENT
        ],
        logger_additional_properties: typing.List[
            logger.property.AbstractProperty
        ] = [],
        output_directory: str = "./",
        folder_name: str = "ioh_data",
        algorithm_name: str = None,
        algorithm_info: str = "algorithm_info",
        store_positions: bool = False,
        experiment_attributes: typing.Dict[str, str] = {},
        run_attributes: typing.List[str] = [],
        logged_attributes: typing.List[str] = [],
        merge_output: bool = True,
        zip_output: bool = True,
        remove_data: bool = False,
        enforce_bounds: bool = False,
        old_logger: bool = True
    ):
        '''
        Parameters
        ----------
            algorithm: typing.Any
                A instance of a given algorithm. Must implement a __call__ method
                taking a ioh.problem as argument.
            fids: typing.List[int]
                A list of integer problem ids to evaluate
            iids: typing.List[int],
                A list of integer problem instances to evaluate
            dims: typing.List[int],
                A list of integer problem dimensions to evaluate
            reps: int = 1,
                The number of independent repetitions for each problem, instance
                dimension combination
            problem_class: ProblemClass = ProblemClass.REAL
                The type of problems to test.
            njobs: int = 1
                The number of parallel jobs, -1 assigns all available cpu's,
            logged: bool = True
                Whether or not the experiment uses a logger
            logger_triggers: typing.List[logger.trigger.Trigger]
                A list of trigger instances, used to determine when to trigger
            logger_additional_properties: typing.List[logger.property.AbstractProperty]
                A list of additional properties to be logged. If the properties to be logged are
                members of `algorithm`, you can also provide a list of attribute names to
                `logged_attributes`.
            output_directory: str = "./"
                The root output directory for the logger
            folder_name: str = "ioh_data"
                The name of the output directory for the logger
            algorithm_name: str = None
                A name for the algorithm. This is used in the log files.
            algorithm_info: str = ""
                Optional information, additional information used in log files
            optimization_type: OptimizationType = OptimizationType.MIN
                The type of optimization
            store_positions: bool = False
                Whether to store the x-positions in the data-files
            experiment_attributes: typing.Dict[str, str] = []
                Attributes additionally stored per experiment. These are static
                throughout the experiment and stored in the .info files.
            run_attributes: typing.List[str] = []
                Names of attributes which are updated at every run, i.e. run index.
                These are stored in the .info files.
                Note that each of these attributes is required to be an attribute
                on algorithm.
            logged_attributes: typing.List[str] = []
                Names of attributes which are updated during the run of the algoritm,
                and are logged in the data files at every time point there is a
                logging action.
                Note that each of these attributes is required to be an attribute
                on algorithm.
            merge_output: bool = True
                Whether to merge output from multiple folders with the same
                folder_name. When run in parallel, seperate folder are produced.
                This option attempts to merge the folders back in single results.
                Note that this will also merge data already present in the folder,
                when the same name is used.
            zip_output: bool = True
                Whether to produce a .zip folder of output
            remove_data: bool = False
                Whether to remove all the produced data, except for the .zip file
                (when produced).
        '''
        if not isinstance(problem_class, ProblemClass):
           raise AttributeError(f"problem_class should be of type {ProblemClass}")

        self.algorithm = algorithm
        self.logger_root = os.path.realpath(os.path.join(output_directory, folder_name))

        self.logger_params = dict(
            triggers=logger_triggers,
            additional_properties=logger_additional_properties,
            root=output_directory,
            folder_name=folder_name,
            algorithm_name=algorithm_name or str(algorithm),
            algorithm_info=algorithm_info,
            store_positions=store_positions,
        )
        self.fids = fids
        self.iids = iids
        self.dims = dims
        self.reps = reps
        self.logged = logged
        self.problem_class = problem_class
        self.njobs = njobs if njobs != -1 else multiprocessing.cpu_count()
        self.experiment_attributes = experiment_attributes
        self.logged_attributes = logged_attributes
        self.run_attributes = run_attributes
        self.merge_output = merge_output
        self.zip_output = zip_output
        self.remove_data = remove_data
        self.enforce_bounds = enforce_bounds
        self.old_logger = old_logger

        if os.path.isdir(self.logger_root) and self.merge_output:
            warnings.warn(
                f"There is already a folder named {self.logger_root} "
                f"and merge_output is set to True. The results of this experiment "
                f"will be merged with the contents of this folder",
                UserWarning,
            )

        for attr in itertools.chain(self.run_attributes, self.logged_attributes):
            if not hasattr(self.algorithm, attr):
                raise TypeError(
                    f"Attribute {attr} is not a member of algorithm {self.algorithm}"
                )


    def evaluate(self, ii: int, job: typing.Tuple[int, int, int]):
        '''Evaluate a single function using self.algoritm.

        Note that this functions makes a copy of the algorithm for each new problem
        instantiation.

        Parameters
        ----------
        ii: int
            Iterator id
        job[fid: int, iid: int, dim: int]
            The problem id, the instance id and the problem dimension
        '''
        algorithm = copy.deepcopy(self.algorithm)
        p = get_problem(*job, self.problem_class)
        if self.logged:
            logger_params = copy.deepcopy(self.logger_params)
            logger_params["folder_name"] += f"-tmp-{ii}"
            
            logger_cls = logger.old.Analyzer if self.old_logger else logger.Analyzer
            l = logger_cls(**logger_params) 
            l.set_experiment_attributes(self.experiment_attributes)
            l.add_run_attributes(algorithm, self.run_attributes)
            l.watch(algorithm, self.logged_attributes)
            l.reset()
            p.attach_logger(l)

        if self.enforce_bounds:
            p.enforce_bounds()

        self.apply(algorithm, p)

    def apply(self, algorithm: any, problem: ProblemType):
        '''Apply a given algorithm to a problem'''
        for _ in range(self.reps):
            algorithm(problem)
            problem.reset()

    def add_custom_problem(self, p: callable, name: str = None, **kwargs):
        '''Add a custom problem to the list of functions to be evaluated.

        Parameters
        ---------
        p: callable
            A function which takes an list of numeric values and return as list
            of numeric values

        name: str
            An optional name of the the newly added function.
        '''

        name = name or "CustomProblem"

        if not self.problem_class.is_real():
            problem.wrap_integer_problem(p, name, **kwargs)
        else:
            problem.wrap_real_problem(p, name, **kwargs)

        self.fids.append(get_problem_id(name, self.problem_class))

    def merge_output_to_single_folder(self, prefix: str, target_folder: str):
        '''Merges all ioh data folders into a single folder, having the same folder name prefix

        Parameters
        ----------
        prexix: str
            The prefix on which to select folders
        target_folder: str
            The target folder, i.e. the folder with the final output
        '''


        def merge_info_file(target, source):                        
            ext = ".info" if self.old_logger else ".json"

            if not (target.endswith(ext) and source.endswith(ext)):
                raise RuntimeError("Merging output with incompatible folders")

            has_file = os.path.isfile(target)
            if not has_file:
                return os.rename(source, target)
            with open(source) as info_in:
                if self.old_logger:
                    with open(target, "a+") as info_out:        
                        info_out.write("\n")
                        for line in info_in:
                            info_out.write(line)
                else:
                    with open(target, "r+") as info_out:
                        data_out = json.loads(info_out.read())
                        data_in = json.loads(info_in.read())
                        for scen in data_in['scenarios']:
                            try:
                                scen_out, *_ = [s for s in data_out['scenarios'] if s['dimension'] == scen['dimension']]
                                scen_out['runs'].extend(scen['runs'])
                            except:
                                data_out['scenarios'].append(scen)
                        info_out.seek(0)
                        info_out.write(json.dumps(data_out, indent=4))
                        info_out.truncate()
            os.remove(source)
                    

        def file_to_dir(path):
            root, dirname = os.path.split(os.path.splitext(path)[0])
            return os.path.join(root, dirname.replace("IOHprofiler", "data"))

        root = os.path.dirname(self.logger_root)

        for folder_name in os.listdir(root):
            folder = os.path.join(root, folder_name)
            if not os.path.isdir(folder):
                continue

            if folder_name.startswith(prefix) and folder_name != os.path.basename(
                target_folder
            ):
                folder = os.path.join(root, folder_name)

                for info_file in os.listdir(folder):
                    source = os.path.join(folder, info_file)
                    if not os.path.isfile(source):
                        continue
                    
                    target = os.path.join(target_folder, info_file)
                    merge_info_file(target, source)

                    source_dat_folder = file_to_dir(source)
                    target_dat_folder = file_to_dir(target)

                    os.makedirs(target_dat_folder, exist_ok=True)
                    for dat_file in os.listdir(source_dat_folder):
                        source_dat_file = os.path.join(source_dat_folder, dat_file)
                        target_dat_file = os.path.join(target_dat_folder, dat_file)

                        with open(source_dat_file) as dat_in, open(
                            target_dat_file, "a+"
                        ) as dat_out:
                            for line in dat_in:
                                dat_out.write(line)

                        os.remove(source_dat_file)
                    os.removedirs(source_dat_folder)

    def merge_tmp_folders(self):
        '''Merges the output of the temporary folders of the experiment'''

        root = os.path.basename(self.logger_root)
        target, idx = self.logger_root, 1
        while os.path.exists(target):
            target = root + f"-{idx}"
            idx += 1

        os.rename(self.logger_root + "-tmp-1", target)
        self.merge_output_to_single_folder(root + "-tmp", target)

    def run(self):
        '''Alias for __call__'''

        return self()

    def __call__(self):
        '''Run the experiment'''

        iterator = enumerate(itertools.product(self.fids, self.iids, self.dims), 1)
        if self.njobs != 1:
            print(self.njobs)
            with multiprocessing.Pool(self.njobs) as p:
                p.starmap(self.evaluate, iterator)
        else:
            for job in iterator:
                self.evaluate(*job)

        self.merge_tmp_folders()

        if self.merge_output:
            self.merge_output_to_single_folder(
                os.path.basename(self.logger_root), self.logger_root
            )

        if self.zip_output:
            shutil.make_archive(self.logger_root, "zip", self.logger_root)

        if self.remove_data:
            shutil.rmtree(self.logger_root)

        return self


__all__ = (
    "get_problem",
    "wrap_problem",
    "get_problem_id",
    "Experiment"
)




from typing import Any, List, Optional, Union, Dict

import ioh.iohcpp as ic

class IntegerSingleObjective:
    problems: Dict[int, str] = ...  # read-only
    def __init__(
        self,
        name: str,
        n_variables: int,
        instance: int,
        is_minimization: bool,
        bounds: ic.IntegerBounds,
        constraints: ic.IntegerConstraint,
        optimum: ic.IntegerSolution,
    ) -> None: ...
    def attach_logger(self, logger) -> None: ...
    def create(
        self, id: Union[int, str], iid: int, dim: int
    ) -> IntegerSingleObjective: ...
    def detach_logger(self) -> None: ...
    def reset(self) -> None: ...
    def __call__(self, x: List[int]) -> float: ...
    def add_constraint(self, constraint: ic.IntegerConstraint) -> None: ...
    def remove_constraint(self, constraint: ic.IntegerConstraint) -> None: ...
    def enforce_bounds(
        self, weight: float, enforced: ic.ConstraintEnforcement, exponent: float
    ) -> None: ...
    @property
    def bounds(self) -> ic.IntegerBounds: ...
    @property
    def constraints(self) -> ic.IntegerConstraintSet: ...
    @property
    def log_info(self) -> ic.ic.LogInfo: ...
    @property
    def meta_data(self) -> ic.MetaData: ...
    @property
    def optimum(self) -> ic.IntegerSolution: ...
    @property
    def state(self) -> ic.IntegerState: ...

class RealSingleObjective:
    problems: Dict[int, str] = ...  # read-only
    def __init__(
        self,
        name: str,
        n_variables: int,
        instance: int,
        is_minimization: bool,
        bounds: ic.RealBounds,
        constraints: ic.RealConstraint,
        optimum: ic.RealSolution,
    ) -> None: ...
    def attach_logger(self, logger) -> None: ...
    def create(
        self, id: Union[int, str], iid: int, dim: int
    ) -> IntegerSingleObjective: ...
    def detach_logger(self) -> None: ...
    def reset(self) -> None: ...
    def __call__(self, x: List[float]) -> float: ...
    def add_constraint(self, constraint: ic.RealConstraint) -> None: ...
    def remove_constraint(self, constraint: ic.RealConstraint) -> None: ...
    def enforce_bounds(
        self, weight: float, enforced: ic.ConstraintEnforcement, exponent: float
    ) -> None: ...
    @property
    def bounds(self) -> ic.RealBounds: ...
    @property
    def constraints(self) -> ic.RealConstraintSet: ...
    @property
    def log_info(self) -> ic.ic.LogInfo: ...
    @property
    def meta_data(self) -> ic.MetaData: ...
    @property
    def optimum(self) -> ic.RealSolution: ...
    @property
    def state(self) -> ic.RealState: ...

class AbstractWModel(IntegerSingleObjective):
    def __init__(
        self,
        problem_id: int,
        instance: int,
        n_variables: int,
        name: str,
        dummy_select_rate: float = ...,
        epistasis_block_size: int = ...,
        neutrality_mu: int = ...,
        ruggedness_gamma: int = ...,
    ) -> None: ...
    def wmodel_evaluate(self, arg0: List[int]) -> int: ...

class AttractiveSector(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class BBOB(RealSingleObjective):
    ...

class BentCigar(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class BuecheRastrigin(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class ConcatenatedTrap(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class DifferentPowers(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class Discus(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class Ellipsoid(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class EllipsoidRotated(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class Gallagher101(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class Gallagher21(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class GraphProblem(IntegerSingleObjective): ...

class GriewankRosenBrock(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class IntegerWrappedProblem(IntegerSingleObjective):
    def __init__(self, *args, **kwargs) -> None: ...

class IsingRing(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class IsingTorus(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class IsingTriangular(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class Katsuura(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class LABS(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class LeadingOnes(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class LeadingOnesDummy1(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class LeadingOnesDummy2(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class LeadingOnesEpistasis(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class LeadingOnesNeutrality(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class LeadingOnesRuggedness1(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class LeadingOnesRuggedness2(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class LeadingOnesRuggedness3(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class Linear(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class LinearSlope(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class LunacekBiRastrigin(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class MIS(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class MaxCoverage(GraphProblem):
    def __init__(self, instance: int, n_variables: int) -> None: ...
    def load_instances(self, path:str) -> None: ...

class MaxCut(GraphProblem):
    def __init__(self, instance: int, n_variables: int) -> None: ...
    def load_instances(self, path:str) -> None: ...

class MaxInfluence(GraphProblem):
    def __init__(self, instance: int, n_variables: int) -> None: ...
    def load_instances(self, path:str) -> None: ...

class NQueens(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class OneMax(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class OneMaxDummy1(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class OneMaxDummy2(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class OneMaxEpistasis(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class OneMaxNeutrality(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class OneMaxRuggedness1(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class OneMaxRuggedness2(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class OneMaxRuggedness3(PBO):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class PBO(IntegerSingleObjective):
    ...

class PackWhileTravel(GraphProblem):
    def __init__(self, *args, **kwargs) -> None: ...
    def load_instances(self, *args, **kwargs) -> Any: ...

class Rastrigin(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class RastriginRotated(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class RealWrappedProblem(RealSingleObjective):
    def __init__(self, *args, **kwargs) -> None: ...

class Rosenbrock(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class RosenbrockRotated(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class Schaffers10(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class Schaffers1000(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class Schwefel(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class SharpRidge(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class Sphere(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class StepEllipsoid(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

class WModelLeadingOnes(AbstractWModel):
    def __init__(
        self,
        instance: int,
        n_variables: int,
        dummy_select_rate: float = ...,
        epistasis_block_size: int = ...,
        neutrality_mu: int = ...,
        ruggedness_gamma: int = ...,
    ) -> None: ...

class WModelOneMax(AbstractWModel):
    def __init__(
        self,
        instance: int,
        n_variables: int,
        dummy_select_rate: float = ...,
        epistasis_block_size: int = ...,
        neutrality_mu: int = ...,
        ruggedness_gamma: int = ...,
    ) -> None: ...

class Weierstrass(BBOB):
    def __init__(self, instance: int, n_variables: int) -> None: ...

def wrap_integer_problem(
    f: callable,
    name: str,
    optimization_type: ic.OptimizationType = ...,
    lb: Optional[float] = ...,
    ub: Optional[float] = ...,
    transform_variables: Optional[callable] = ...,
    transform_objectives: Optional[callable] = ...,
    calculate_objective: Optional[callable] = ...,
) -> None: ...
def wrap_real_problem(
    f: callable,
    name: str,
    optimization_type: ic.OptimizationType = ...,
    lb: Optional[float] = ...,
    ub: Optional[float] = ...,
    transform_variables: Optional[callable] = ...,
    transform_objectives: Optional[callable] = ...,
    calculate_objective: Optional[callable] = ...,
) -> None: ...

class RealStarDiscrepancy(RealSingleObjective):
    def __init__(
        self,
        instance: int,
        n_variables: int,
        n_samples: int,
        sampler_type: ic.StarDiscrepancySampler,
    ) -> None: ...

class IntegerStarDiscrepancy(IntegerSingleObjective):
    def __init__(
        self,
        instance: int,
        n_variables: int,
        n_samples: int,
        sampler_type: ic.StarDiscrepancySampler,
    ) -> None: ...



from typing import Dict, List, Optional, Tuple
from . import eaf, eah, property, trigger

from typing import overload
import ioh.iohcpp

class AbstractWatcher(Logger):
    def watch(self, arg0: property.AbstractProperty) -> None: ...

class AbstractLogger(Logger):
    def __init__(self, triggers: List[trigger.Trigger], properties: List[property.AbstractProperty]) -> None: ... 
    def watch(self, arg0: property.AbstractProperty) -> None: ...

class Analyzer(AbstractWatcher):
    def __init__(self, triggers: List[trigger.Trigger] = ..., additional_properties: List[property.AbstractProperty] = ..., root: Path = ..., folder_name: str = ..., algorithm_name: str = ..., algorithm_info: str = ..., store_positions: bool = ...) -> None: ...
    def add_experiment_attribute(self, arg0: str, arg1: str) -> None: ...
    @overload
    def add_run_attributes(self, arg0: str, arg1: float) -> None: ...
    @overload
    def add_run_attributes(self, arg0: object, arg1: str) -> None: ...
    @overload
    def add_run_attributes(self, arg0: object, arg1: List[str]) -> None: ...
    def set_experiment_attributes(self, arg0: Dict[str,str]) -> None: ...
    def set_run_attribute(self, arg0: str, arg1: float) -> None: ...
    def set_run_attributes(self, arg0: Dict[str,float]) -> None: ...
    @overload
    def watch(self, arg0: property.AbstractProperty) -> None: ...
    @overload
    def watch(self, arg0: object, arg1: str) -> None: ...
    @overload
    def watch(self, arg0: object, arg1: List[str]) -> None: ...
    @property
    def output_directory(self) -> Path: ...

class Combine(Logger):
    @overload
    def __init__(self, loggers: List[Logger]) -> None: ...
    @overload
    def __init__(self, logger: Logger) -> None: ...
    def append(self, arg0: Logger) -> None: ...

class EAF(Logger):
    def __init__(self) -> None: ...
    def at(self, arg0: str, arg1: int, arg2: int, arg3: int, arg4: int) -> List[eaf.RunPoint]: ...
    @property
    def data(self) -> Dict[str,Dict[int,Dict[int,Dict[int,Dict[int,List[eaf.RunPoint]]]]]]: ...

class EAH(Logger):
    @overload
    def __init__(self, error_min: float, error_max: float, error_buckets: int, evals_min: int, evals_max: int, evals_buckets: int) -> None: ...
    @overload
    def __init__(self, error_scale: eah.LinearRealScale, eval_scale: eah.LinearIntegerScale) -> None: ...
    @overload
    def __init__(self, error_scale: eah.Log2RealScale, eval_scale: eah.Log2IntegerScale) -> None: ...
    @overload
    def __init__(self, error_scale: eah.Log10RealScale, eval_scale: eah.Log10IntegerScale) -> None: ...
    def at(self, arg0: int, arg1: int, arg2: int, arg3: int) -> List[List[bool]]: ...
    @property
    def data(self) -> Dict[int,Dict[int,Dict[int,Dict[int,List[List[bool]]]]]]: ...
    @property
    def error_range(self) -> eah.RealScale: ...
    @property
    def eval_range(self) -> eah.IntegerScale: ...
    @property
    def size(self) -> Tuple[int,int,int,int]: ...

class FlatFile(AbstractWatcher):
    def __init__(self, triggers: List[trigger.Trigger], properties: List[property.AbstractProperty], filename: str = ..., output_directory: Path = ..., separator: str = ..., comment: str = ..., no_value: str = ..., end_of_line: str = ..., repeat_header: bool = ..., store_positions: bool = ..., common_header_titles: List[str] = ...) -> None: ...
    @overload
    def watch(self, arg0: property.AbstractProperty) -> None: ...
    @overload
    def watch(self, arg0: object, arg1: str) -> None: ...
    @overload
    def watch(self, arg0: object, arg1: List[str]) -> None: ...
    @property
    def filename(self) -> str: ...
    @property
    def output_directory(self) -> str: ...

class Logger:
    def __init__(self, *args, **kwargs) -> None: ...
    def add_trigger(self, arg0: trigger.Trigger) -> None: ...
    def attach_problem(self, arg0: ioh.iohcpp.MetaData) -> None: ...
    def attach_suite(self, arg0: str) -> None: ...
    def call(self, arg0: ioh.iohcpp.LogInfo) -> None: ...
    def log(self, arg0: ioh.iohcpp.LogInfo) -> None: ...
    def reset(self) -> None: ...
    @property
    def problem(self) -> ioh.iohcpp.MetaData: ...

class Path:
    def __init__(self, arg0: str) -> None: ...

class Store(AbstractWatcher):
    def __init__(self, triggers: List[trigger.Trigger], properties: List[property.AbstractProperty]) -> None: ...
    def at(self, arg0: str, arg1: int, arg2: int, arg3: int, arg4: int, arg5: int) -> Dict[str,Optional[float]]: ...
    def data(self) -> Dict[str,Dict[int,Dict[int,Dict[int,Dict[int,Dict[int,Dict[str,Optional[float]]]]]]]]: ...
    @overload
    def watch(self, arg0: property.AbstractProperty) -> None: ...
    @overload
    def watch(self, arg0: object, arg1: str) -> None: ...
    @overload
    def watch(self, arg0: object, arg1: List[str]) -> None: ...



import typing
import enum
from .iohcpp import IntegerSolution, OptimizationType, RealSolution, RealBounds, RealConstraint, IntegerBounds, IntegerConstraint

from .iohcpp import problem as problem
from .iohcpp import suite as suite
from .iohcpp import logger as logger

from typing import Any

ProblemType = typing.Union[problem.RealSingleObjective, problem.IntegerSingleObjective]
VariableType = typing.Union[int, float]
ObjectiveType = typing.List[VariableType]

class ProblemClass(enum.Enum):
    """Enum for different problem types, values are class names"""

    REAL = "RealSingleObjective"
    INTEGER = "IntegerSingleObjective"

    BBOB = "BBOB"
    STAR_REAL = "RealStarDiscrepancy"
    SBOX = "SBOX"

    PBO = "PBO"
    GRAPH = "GraphProblem"
    STAR_INTEGER = "IntegerStarDiscrepancy"

    def is_real(self):
        ...

    def is_single_objective(self): 
        ...

def get_problem(fid: typing.Union[int, str], instance: int = ..., dimension: int = ..., problem_type: ProblemClass = ...) -> ProblemType: ...
def wrap_problem(function: typing.Callable[[ObjectiveType], float], name: str, problem_type: ProblemClass, dimension: int = ..., instance: int = ..., optimization_type: OptimizationType = ..., lb: VariableType = ..., ub: VariableType = ..., transform_variables: typing.Callable[[ObjectiveType, int], ObjectiveType] = ..., transform_objectives: typing.Callable[[float, int], float] = ..., calculate_objective: typing.Callable[[int, int], typing.Union[IntegerSolution, RealSolution]] = ...) -> ProblemType: ...
def get_problem_id(problem_name: str, problem_type: ProblemClass) -> int: ...

class Experiment:
    algorithm: Any
    logger_root: str
    logger_params: Any
    fids: typing.List[int]
    iids: typing.List[int]
    dims: typing.List[int]
    reps: int
    logged: bool
    problem_type: ProblemClass
    njobs: int
    experiment_attributes: typing.Dict
    logged_attributes: Any
    run_attributes: Any
    merge_output: bool
    zip_output: bool
    remove_data: bool
    def __init__(self, algorithm: typing.Any, fids: typing.List[int], iids: typing.List[int], dims: typing.List[int], reps: int = ..., problem_type: str = ..., njobs: int = ..., logged: bool = ..., logger_triggers: typing.List[logger.trigger.Trigger] = ..., logger_additional_properties: typing.List[logger.property.AbstractProperty] = ..., output_directory: str = ..., folder_name: str = ..., algorithm_name: str = ..., algorithm_info: str = ..., store_positions: bool = ..., experiment_attributes: typing.Dict[str, str] = ..., run_attributes: typing.List[str] = ..., logged_attributes: typing.List[str] = ..., merge_output: bool = ..., zip_output: bool = ..., remove_data: bool = ...) -> None: ...
    def evaluate(self, ii: int, job: typing.Tuple[int, int, int]): ...
    def apply(self, algorithm: any, problem: ProblemType): ...
    def add_custom_problem(self, p: callable, name: str = None, **kwargs): ...
    def merge_output_to_single_folder(self, prefix: str, target_folder: str): ...
    def merge_tmp_folders(self) -> None: ...
    def run(self): ...
    def __call__(self): ...



from typing import Optional

import ioh.iohcpp
CURRENT_Y_BEST: CurrentY
EVALUATIONS: Evaluations
RAW_Y_BEST: RawYBest
TRANSFORMED_Y: TransformedY
TRANSFORMED_Y_BEST: TransformedYBest

class AbstractProperty:
    def __init__(self, *args, **kwargs) -> None: ...
    def call_to_string(self, arg0: ioh.iohcpp.LogInfo, arg1: str) -> str: ...
    def name(self) -> str: ...
    def __call__(self, arg0: ioh.iohcpp.LogInfo) -> Optional[float]: ...

class CurrentY(AbstractProperty):
    def __init__(self, name: str, format: str) -> None: ...
    def __getstate__(self) -> tuple: ...
    def __setstate__(self, arg0: tuple) -> None: ...

class Evaluations(AbstractProperty):
    def __init__(self, name: str, format: str) -> None: ...
    def __getstate__(self) -> tuple: ...
    def __setstate__(self, arg0: tuple) -> None: ...

class Property(AbstractProperty):
    def __init__(self, container: object, attribute: str) -> None: ...
    def __getstate__(self) -> tuple: ...
    def __setstate__(self, arg0: tuple) -> None: ...

class RawYBest(AbstractProperty):
    def __init__(self, name: str, format: str) -> None: ...
    def __getstate__(self) -> tuple: ...
    def __setstate__(self, arg0: tuple) -> None: ...

class TransformedY(AbstractProperty):
    def __init__(self, name: str, format: str) -> None: ...
    def __getstate__(self) -> tuple: ...
    def __setstate__(self, arg0: tuple) -> None: ...

class TransformedYBest(AbstractProperty):
    def __init__(self, name: str, format: str) -> None: ...
    def __getstate__(self) -> tuple: ...
    def __setstate__(self, arg0: tuple) -> None: ...



import os
import sys
import venv
import glob
import tempfile
import unittest
import subprocess

DIR = os.path.realpath(os.path.join(__file__, "..", "..", ".."))


class TestPackage(unittest.TestCase):
    """Test if package can be installed (takes a looong time)."""

    @unittest.skipUnless(
        sys.platform != "win32" and sys.version_info.minor >= 8,
        "pip location on windows",
    )
    def test_can_install_source_dist(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            result = subprocess.run(
                [
                    sys.executable,
                    os.path.join(DIR, "setup.py"),
                    "sdist",
                    f"--dist-dir={tmpdirname}",
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            self.assertEqual(result.returncode, 0, msg=result.stderr)

            source_dist, *_ = glob.glob(f"{tmpdirname}/*")
            self.assertTrue(os.path.isfile(source_dist))

            venv_dir = os.path.join(tmpdirname, "venv")
            venv.create(venv_dir, with_pip=True)

            subprocess.check_call(
                [os.path.join(venv_dir, "bin", "pip"), "install", source_dist],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )


if __name__ == "__main__":
    unittest.main()



import unittest
import ioh
import os
import shutil
import pickle
import json


class Container:
    def __init__(self):
        self.xy = 0
        self.xv = 10


class TestLogger(unittest.TestCase):
    def test_flatfile(self):
        pr = ioh.get_problem("Sphere", 1, 5) 
        p = [ioh.logger.property.CURRENTBESTY]
        t = [ioh.logger.trigger.ON_IMPROVEMENT]
        l = ioh.logger.FlatFile(t, p, separator=" ")
        pr.attach_logger(l)
        self.assertEqual(os.path.realpath("./"), os.path.realpath(l.output_directory))
        pr([0] * 5)
        self.assertTrue(os.path.isfile("IOH.dat"))

    def test_analyzer(self):
        pr = ioh.get_problem("Sphere", 1, 5) 
        c = Container()
        l = ioh.logger.Analyzer([ioh.logger.trigger.ALWAYS])    
        l.set_experiment_attributes({"x":"1"})
        l.watch(c, ["xv", "xy"])
        l.add_run_attribute(c, "xv")
        pr.attach_logger(l)

        for c.xv in range(4):
            for i in range(5):
                c.xy = i*10
                pr([i] * 5)
            pr.reset()
        l.close()
        
        self.assertTrue(os.path.isdir("ioh_data"))
        meta = os.path.join("ioh_data", "IOHprofiler_f1_Sphere.json")
        data = os.path.join("ioh_data", "data_f1_Sphere", "IOHprofiler_f1_DIM5.dat")
        self.assertTrue(os.path.isfile(meta))
        self.assertTrue(os.path.isfile(data))

        with open(meta) as f:
            meta_data = json.loads(f.read())
            for i, run in enumerate(meta_data["scenarios"][0]["runs"]):
                self.assertEqual(run['xv'], i)
        with open(data) as f:
            next(f)
            for i in range(5):
                self.assertEqual(i*10, float(next(f).split()[-1]))

    def test_triggers(self):
        always = ioh.logger.trigger.Always()
        on_improvement = ioh.logger.trigger.OnImprovement()
        at = ioh.logger.trigger.At({1, 22})
        self.assertSetEqual(at.time_points, {1, 22})

        each = ioh.logger.trigger.Each(10, 2)
        self.assertEqual(each.interval, 10)
        self.assertEqual(each.starting_at, 2)

        during = ioh.logger.trigger.During({(10, 25), (30, 40)})
        self.assertEqual(during.time_ranges, {(10, 25), (30, 40)})

        l = ioh.LogInfo(22, 1, 1, 1, 1, 1, 1, [0, 1],[0, 1],[0, 1], ioh.RealSolution([0, 1], 0), True)
        problem = ioh.get_problem(1, 1, 2)
        for t in (always, on_improvement, at, each, during):
            self.assertTrue(t(l, problem.meta_data))


    def test_multiple_problems_in_sequence(self):
        p1 = ioh.get_problem(1, 2, 5)
        p2 = ioh.get_problem(2, 2, 5)
        l = ioh.logger.Analyzer() 
        for i in range(5):
            for p in p1, p2:
                p.attach_logger(l)
                for _ in range(10):
                    p([0] * 5)
                p.reset()
        
        for p in p1, p2:
            i = p.meta_data.problem_id
            name = p.meta_data.name
            meta = os.path.realpath(os.path.join("ioh_data", f"IOHprofiler_f{i}_{name}.json"))
            data = os.path.realpath(os.path.join("ioh_data", f"data_f{i}_{name}", f"IOHprofiler_f{i}_DIM5.dat"))
            self.assertTrue(os.path.isfile(meta))
            self.assertTrue(os.path.isfile(data))
            with open(meta) as f:
                meta_data = json.load(f)
                self.assertEqual(len(meta_data["scenarios"][0]["runs"]), 5)
            with open(data) as f:
                n_headers = 0
                for line in f:
                    if line.startswith("evaluations"):
                        n_headers += 1
                self.assertEqual(n_headers, 5)


    def test_multiple_problems_has_info_file(self):
        p1 = ioh.get_problem(1, 2, 5)
        p2 = ioh.get_problem(2, 2, 5)
        l = ioh.logger.Analyzer() 


    def tearDown(self) -> None:
        if os.path.isdir("ioh_data"):
            shutil.rmtree("ioh_data")
        if os.path.isfile("IOH.dat"):
            os.remove("IOH.dat")
        return super().tearDown()

    def test_pickle(self):
        try:
            pickle.dumps(ioh.logger.trigger.ON_IMPROVEMENT)
        except Exception as err: 
            self.fail(str(err))
        try:
            pickle.dumps(ioh.logger.property.RAWYBEST)
        except Exception as err: 
            self.fail(str(err))
        try:
            pickle.dumps(ioh.logger.property.Property(Container(), "xy"))
        except Exception as err: 
            self.fail(str(err))

    def test_eaf(self):
        p = ioh.get_problem(1, 1, 5)
        l = ioh.logger.EAH(0, 20, 10, 0, 100, 10)
        self.assertEqual(l.error_range.min, 0)
        self.assertEqual(l.error_range.max, 20)
        self.assertEqual(l.error_range.size, 10)
        self.assertEqual(l.eval_range.min, 0)
        self.assertEqual(l.eval_range.max, 100)
        self.assertEqual(l.eval_range.size, 10)
        p.attach_logger(l)

        for i in range(10):
            p([i] * 5)

        self.assertEqual(list(l.data.keys()), [1])
                    
    def test_can_inherit_property(self):
        class P(ioh.logger.property.AbstractProperty):
            def __call__(self, _):
                return 10.0
        c = ioh.RealSolution([0], 1)
        l = ioh.LogInfo(0, 1, 1, 1, 1, 1, 1, [0],[0],[0], c)
        a = P("x")
        self.assertEqual(int(a(l)), int(float(a.call_to_string(l, "na"))))
        

if __name__ == "__main__":
    unittest.main()


# -*- coding: utf-8 -*-
import os
import shutil
import sys
import platform
import subprocess
import atexit
from setuptools import setup, Extension, find_packages 
from setuptools.command.build_ext import build_ext

from doc.generate_docs import main, generate_stubs



DIR = os.path.realpath(os.path.dirname(__file__))
with open(os.path.join(DIR, "VERSION")) as f:
    __version__ = f.read().strip()

BUILD_DIR = os.path.join(DIR, "build")
if os.path.exists(BUILD_DIR):
    shutil.rmtree(BUILD_DIR)

# Convert distutils Windows platform specifiers to CMake -A arguments
PLAT_TO_CMAKE = {
    "win32": "Win32",
    "win-amd64": "x64",
    "win-arm32": "ARM",
    "win-arm64": "ARM64",
}

if platform.system() == "Darwin":
    os.environ["CC"] = "clang"
    os.environ["CXX"] = "clang"
    os.environ["ARCHFLAGS"] = "-std=c++14"

BASE_DIR = os.path.realpath(os.path.dirname(__file__))
MAKE_DOCS = os.environ.get("MAKE_DOCS")
MAKE_STUBS = os.environ.get("MAKE_STUBS")

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        super().run()
        if MAKE_STUBS:
            ext, *_ = self.extensions
            generate_stubs(ext.sourcedir)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

        # required for auto-detection of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        cfg = "Debug" if self.debug else "Release"
        print("build-type:", cfg)

        # CMake lets you override the generator - we need to check this.
        # Can be set with Conda-Build, for example.
        cmake_generator = os.environ.get("CMAKE_GENERATOR", "")

        # Set Python_EXECUTABLE instead if you use PYBIND11_FINDPYTHON
        # EXAMPLE_VERSION_INFO shows you how to pass a value into the C++ code
        # from Python.
        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={}".format(extdir),
            "-DPYTHON_EXECUTABLE={}".format(sys.executable),
            "-DBUILD_PYTHON_PACKAGE=ON",
            "-DBUILD_TESTS=OFF",
            "-DENABLE_MKLANDSCAPE_PROBLEMS=OFF",
            "-DBUILD_DOCS={}".format("ON" if MAKE_DOCS else "OFF"),
            "-DBUILD_EXAMPLE=OFF",
            "-DCMAKE_BUILD_TYPE={}".format(cfg),  # not used on MSVC, but no harm
        ]
        build_args = []

        if self.compiler.compiler_type != "msvc":
            # Using Ninja-build since it a) is available as a wheel and b)
            # multithreads automatically. MSVC would require all variables be
            # exported for Ninja to pick it up, which is a little tricky to do.
            # Users can override the generator with CMAKE_GENERATOR in CMake
            # 3.15+.
            if not cmake_generator:
                cmake_args += ["-GNinja"]

        else:

            # Single config generators are handled "normally"
            single_config = any(x in cmake_generator for x in {"NMake", "Ninja"})

            # CMake allows an arch-in-generator style for backward compatibility
            contains_arch = any(x in cmake_generator for x in {"ARM", "Win64"})

            # Specify the arch if using MSVC generator, but only if it doesn't
            # contain a backward-compatibility arch spec already in the
            # generator name.
            if not single_config and not contains_arch:
                cmake_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]

            # Multi-config generators have a different way to specify configs
            if not single_config:
                cmake_args += [
                    "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)
                ]
                build_args += ["--config", cfg]

        # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
        # across all generators.
        if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
            # self.parallel is a Python 3 only way to set parallel jobs by hand
            # using -j in the build_ext call, not supported by pip or PyPA-build.
            if hasattr(self, "parallel") and self.parallel:
                # CMake 3.12+ only.
                build_args += ["-j{}".format(self.parallel)]

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        subprocess.check_call(
            ["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp
        )
        subprocess.check_call(
            ["cmake", "--build", "."] + build_args, cwd=self.build_temp
        )


if MAKE_DOCS:
    try:
        atexit.register(main)
    except:
        pass

with open("README.md", "r") as fh:
    long_description = fh.read()

iohcpp = CMakeExtension("ioh.iohcpp")

setup(
    name="ioh",
    version=__version__,
    author="Jacob de Nobel, Furong Ye, Diederick Vermetten, Hao Wang, Carola Doerr and Thomas Bäck",
    author_email="iohprofiler@liacs.leidenuniv.nl",
    description="The experimenter for Iterative Optimization Heuristics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    package_dir={"IOHexperimenter": "ioh"},
    include_package_data=True,
    exclude_package_data={"": ["src/*", "*CMakeLists.txt", "*README.md"]},
    ext_modules=[iohcpp],
    cmdclass={"build_ext": CMakeBuild},
    zip_safe=False,
    test_suite="tests.python",
    python_requires=">=3.6",
    setup_requires=[
        "setuptools",
        "wheel",
        "cmake",
        "mypy",
        "pybind11",
        "ninja",
        "xmltodict",
    ],
    license="BSD",
    url="https://iohprofiler.github.io/IOHexperimenter",
    install_requires=["numpy"]
)



from typing import Tuple

class IntegerScale:
    def __init__(self, *args, **kwargs) -> None: ...
    def bounds(self, arg0: int) -> Tuple[float,float]: ...
    def index(self, arg0: float) -> int: ...
    @property
    def length(self) -> int: ...
    @property
    def max(self) -> int: ...
    @property
    def min(self) -> int: ...
    @property
    def size(self) -> int: ...

class LinearIntegerScale(IntegerScale):
    def __init__(self, arg0: int, arg1: int, arg2: int) -> None: ...
    def step(self) -> int: ...

class LinearRealScale(RealScale):
    def __init__(self, arg0: float, arg1: float, arg2: int) -> None: ...
    def step(self) -> float: ...

class Log10IntegerScale(IntegerScale):
    def __init__(self, arg0: int, arg1: int, arg2: int) -> None: ...

class Log10RealScale(RealScale):
    def __init__(self, arg0: float, arg1: float, arg2: int) -> None: ...

class Log2IntegerScale(IntegerScale):
    def __init__(self, arg0: int, arg1: int, arg2: int) -> None: ...

class Log2RealScale(RealScale):
    def __init__(self, arg0: float, arg1: float, arg2: int) -> None: ...

class RealScale:
    def __init__(self, *args, **kwargs) -> None: ...
    def bounds(self, arg0: int) -> Tuple[float,float]: ...
    def index(self, arg0: float) -> int: ...
    @property
    def length(self) -> float: ...
    @property
    def max(self) -> float: ...
    @property
    def min(self) -> float: ...
    @property
    def size(self) -> int: ...



import os
import unittest
import math

import ioh
import numpy as np


DATA_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(__file__))), "static"
)


class wmodel(ioh.problem.AbstractWModel):
    def __init__(self, instance, dim):
        super().__init__(100, instance, dim, "wmodel")

    def wmodel_evaluate(self, x) -> int:
        return x.count(1)


class TestProblem(unittest.TestCase):
    def test_get_problem(self):
        self.assertIsInstance(
            ioh.get_problem(1, 1, 2, ioh.ProblemClass.BBOB), ioh.problem.Sphere
        )
        self.assertIsInstance(
            ioh.get_problem("Sphere", 1, 2, ioh.ProblemClass.BBOB), ioh.problem.Sphere
        )
        self.assertIsInstance(ioh.get_problem("Sphere", 1, 2), ioh.problem.Sphere)
        self.assertIsInstance(
            ioh.get_problem(1, 1, 2, ioh.ProblemClass.PBO), ioh.problem.OneMax
        )
        self.assertIsInstance(
            ioh.get_problem("OneMax", 1, 2, ioh.ProblemClass.PBO), ioh.problem.OneMax
        )

    def test_vectorized_eval(self):
        n_samples = 5
        dim = 2

        problem = ioh.get_problem(1, 1, dim)
        X = np.zeros((n_samples, dim))
        y = problem(X)
        self.assertIsInstance(y, list)
        self.assertEqual(len(y), n_samples)
        self.assertEqual(problem.state.evaluations, n_samples)
        self.assertSetEqual(set(y), {80.88209408})

    def test_wmodel(self):
        for p in map(
            lambda x: x(1, 10),
            (wmodel, ioh.problem.WModelLeadingOnes, ioh.problem.WModelOneMax),
        ):
            self.assertEqual(p([1] * 10), 10)

    def test_can_access_optimization_type(self): 
        p = ioh.get_problem(1, 1, 2)
        self.assertTrue(p.meta_data.optimization_type is not None)

    def test_class_based_problem(self):
        class A(ioh.problem.RealSingleObjective):
            def __init__(self, instance, dim) -> None:
                super().__init__(
                    "A",
                    dim,
                    instance,
                    False,
                    ioh.RealBounds(dim, 10, 20),
                    [], 
                    ioh.RealSolution([1] * dim, dim)
                )
        a = A(1, 2)
        try:
            a([0] * 2)
            self.fail()
        except RuntimeError:
            pass

        setattr(a, "evaluate", lambda x:sum(xi == 1 for xi in x))
        self.assertEqual(a([0, 0]), 0)
        self.assertEqual(a([0, 1]), 1)
        self.assertFalse(a.state.optimum_found)
        self.assertEqual(a([1, 1]), 2)
        self.assertTrue(a.state.optimum_found)

    def test_real_star_discrepancy(self):
        uniform = ioh.get_problem(
            30,
            4,
            2,
        )
        self.assertAlmostEqual(uniform([0.9, 0.5]), 0.25, 3)
        sobol = ioh.get_problem(40, 4, 2)
        self.assertAlmostEqual(sobol([0.9, 0.5]), 0.0499, 3)
        halton = ioh.get_problem(50, 4, 2)
        self.assertAlmostEqual(halton([0.9, 0.5]), 0.1499, 3)

        random = ioh.problem.RealStarDiscrepancy(
            instance=69,
            n_variables=2,
            n_samples=9,
            sampler_type=ioh.problem.StarDiscrepancySampler.HALTON,
        )
        self.assertEqual(random([10, 10]), -float("inf"))

        self.assertAlmostEqual(random([0.9, 0.5]), 0.0055, 3)

        self.assertTrue(math.isnan(uniform.optimum.x[0]))
        self.assertTrue(math.isinf(uniform.optimum.y))

    def test_integer_star_discrepancy(self):
        uniform = ioh.get_problem(30, 4, 2, problem_class=ioh.ProblemClass.STAR_INTEGER)
        self.assertAlmostEqual(uniform([9, 9]), 0.11825, 3)

        sobol = ioh.get_problem(40, 4, 2, problem_class=ioh.ProblemClass.STAR_INTEGER)
        self.assertAlmostEqual(sobol([9, 9]), 0.1797, 3)

        halton = ioh.get_problem(50, 4, 2, problem_class=ioh.ProblemClass.STAR_INTEGER)
        self.assertAlmostEqual(halton([9, 9]), 0.2222, 3)

        random = ioh.problem.IntegerStarDiscrepancy(
            instance=69,
            n_variables=2,
            n_samples=10,
            sampler_type=ioh.problem.StarDiscrepancySampler.HALTON,
        )
        self.assertEqual(random([10, 10]), 0)
        self.assertEqual(random([11, 11]), -float("inf"))

    def test_enforced_bounds(self):
        p = ioh.get_problem(1, 1, 2)
        x0 = [10, 1]
        y0 = p(x0)
        p.enforce_bounds()
        y1 = p(x0)
        self.assertEqual(y0 + p.constraints.penalty(), y1)

    def test_custom_constraint(self):
        p = ioh.get_problem(1, 1, 2)
        c = ioh.RealConstraint(lambda x: float(x[0] > 1), 25.0)
        y0 = p([10, 1])
        p.add_constraint(c)
        self.assertEqual(p([10, 1]), y0 + 25)

    def test_evaluation_bbob_problems(self):
        for fid in range(1, 25):
            f = ioh.get_problem(fid, 1, 5, ioh.ProblemClass.BBOB)
            self.assertGreater(f([0, 0, 0, 0, 0]), -1000)

    def test_evaluation_pbo_problems(self):
        for fid in range(1, 26):
            f = ioh.get_problem(fid, 1, 4, ioh.ProblemClass.PBO)
            self.assertGreater(f([0, 0, 0, 0]), -1000)

    def test_has_submodular(self):
        problems = list(getattr(ioh.problem, "GraphProblem").problems.values())
        self.assertNotEqual(0, len(problems))
        for name in ("MaxCut", "MaxInfluence", "MaxCoverage", "PackWhileTravel"):
            self.assertNotEqual(0, len([p for p in problems if name in p]))

    def test_bbob_problems_first_instance(self):
        expected = [
            161.17445568,
            12653420.412225708,
            -97.06158297486468,
            8845.53524810093,
            121.9133638725643,
            423021.00682286796,
            278.3290201933729,
            4315.032550201522,
            95168.253629631741,
            969025.0451803299,
            20911310.769634742,
            187251637.66430587,
            2198.7155122256763,
            261.60997479957,
            1103.1553337856833,
            273.87870697835791,
            35.410739487431208,
            107.64134358767826,
            -40.269198932753994,
            49739.05388887795,
            116.29585727504872,
            -914.902473409051,
            18.635078550302751,
            1782.2733296400438,
        ]
        for i in sorted(ioh.problem.BBOB.problems.keys())[:24]:
            p = ioh.problem.BBOB.create(i, 1, 5)
            self.assertTrue(math.isclose(p([0.1, 1.0, 2.0, 4.0, 5.4]), expected[i - 1]))

    def test_pbo_problems_first_instance(self):
        expected = [
            5.0000,
            2.0000,
            24.0000,
            3.0000,
            4.0000,
            2.0000,
            6.0000,
            4.0000,
            6.0000,
            7.0000,
            3.0000,
            2.0000,
            1.0000,
            0.0000,
            2.0000,
            1.0000,
            1.0000,
            2.5312,
            5.0000,
            6.0000,
            9.0000,
            -4.0000,
            -16.0000,
            0.45,
            -0.70717,
        ]
        for i in sorted(ioh.problem.PBO.problems.keys())[:24]:
            p = ioh.problem.PBO.create(i, 1, 9)
            y = p([1, 1, 0, 1, 0, 0, 0, 1, 1])
            self.assertTrue(
                math.isclose(y, expected[i - 1], abs_tol=0.000099),
                msg=f"{p} expected: {expected[i-1]} got: {y}",
            )

    def test_file_comparisons(self):
        for test_file in (
            "pbofitness16.in",
            "pbofitness100.in",
            "bbobfitness5.in",
            "bbobfitness20.in",
        ):
            with self.subTest(test_file=test_file):
                suite, dim = test_file.split("fitness")
                problem_class = (
                    ioh.ProblemClass.PBO if suite == "pbo" else ioh.ProblemClass.BBOB
                )
                dim = int(dim[:-3])
                dtype = float if suite == "bbob" else int
                tol = 0.01 if suite == "bbob" else 0.000099

                with open(os.path.join(DATA_DIR, test_file)) as f:
                    for line in f:
                        fid, iid, x, y = line.split()
                        if "," in x:
                            x = x.split(",")
                        x = list(map(dtype, x))
                        p = ioh.get_problem(int(fid), int(iid), dim, problem_class)
                        self.assertTrue(math.isclose(p(x), float(y), abs_tol=tol))

    def test_sbox(self):
        for pid, name in ioh.problem.SBOX.problems.items():
            sbox = ioh.get_problem(pid, 1, 4, ioh.ProblemClass.SBOX)
            bbob = ioh.get_problem(pid, 1, 4, ioh.ProblemClass.BBOB)
            self.assertEqual(sbox.constraints.n(), 1)
            self.assertEqual(sbox([10, 10, 10, 10]), float("inf"))
            self.assertNotEqual(bbob([10, 10, 10, 10]), float("inf"))
            self.assertAlmostEqual(bbob.optimum.y, sbox.optimum.y)

            if pid in (5, 9, 19, 20):
                continue
            self.assertFalse(all(bbob.optimum.x == sbox.optimum.y))

    def test_sbox_suite(self):
        suite = ioh.suite.SBOX(range(25))
        self.assertEqual(len(suite), 24)


if __name__ == "__main__":
    unittest.main()
